mod readmap_helper;
use crossterm::event::KeyCode;
use readmap::types::MemoryMap;
use std::path::PathBuf;
use tool_helper::Error;
use ratatui::{
    layout::{Alignment, Constraint, Direction, Layout, Rect},
    style::{Color, Modifier, Style},
    text::{Line, Span},
    widgets::{Block, Borders, BorderType, Gauge, List, ListItem, ListState, Paragraph, Tabs}
};

pub type EventReceiver = std::sync::mpsc::Receiver<Event<crossterm::event::KeyEvent>>;
pub type Terminal      = ratatui::Terminal<ratatui::backend::CrosstermBackend<std::io::Stdout>>;
type ConsoleFrame<'a>  = ratatui::Frame<'a, ratatui::backend::CrosstermBackend<std::io::Stdout>>;

pub enum Event<I> {
    Input(I),
    Tick,
    ForceRender,
}

pub enum UIState {
    Alive,
    Render,
    Terminated
}

pub enum MenuSelection {
    Stats,
    Quit,
}

impl MenuSelection {
    fn next(&self) -> MenuSelection {
        match self {
            Self::Stats => Self::Quit,
            Self::Quit  => Self::Stats
        }
    }

    fn prev(&self) -> MenuSelection {
        self.next()
    }

    fn get_id(&self) -> usize {
        match self {
            MenuSelection::Stats => 0,
            MenuSelection::Quit  => 1,
        }
    }
}

pub struct ListSelection {
    pub selection: Option<usize>
}

impl ListSelection {
    fn change_value(&mut self, max: usize, f: fn(usize)->usize) {
        if let Some(cur_selection) = self.selection {
            self.selection = Some(f(cur_selection)%max);
        }
    }

    pub fn increment(&mut self, max: usize) {
        fn add_one(value: usize) -> usize {
            value + 1
        }
        self.change_value(max, add_one)
    }

    pub fn decrement(&mut self, max: usize) {
        fn sub_one(value: usize) -> usize {
            value - 1
        }
        self.change_value(max, sub_one)
    }

    pub fn activate(&mut self) {
        self.selection = Some(0);
    }

    pub fn deactivate(&mut self) {
        self.selection = None;
    }

    pub fn is_active(&self) -> bool {
        self.selection.is_some()
    }
}

impl Default for ListSelection {
    fn default() -> Self {
        Self{selection: None}
    }
}

pub struct ConsoleUI {
    data:              ConsoleUIData,
    rx:                EventReceiver,
    terminal:          Terminal,
    console_size:      Rect,
    menu_selection:    MenuSelection,
    content_selection: ListSelection,
}

impl ConsoleUI {
    pub fn new(rx: EventReceiver, terminal: Terminal) -> ConsoleUI {
        ConsoleUI{data: ConsoleUIData::default(), rx, terminal, console_size: Rect::default(), menu_selection: MenuSelection::Stats, content_selection: ListSelection::default()}
    }

    pub fn update_data(&mut self, memory_map: &MemoryMap) -> Result<(), Error> {
        self.data.update(memory_map)?;

        if self.data.list.len() > 0 {
            if !self.content_selection.is_active() {
                self.content_selection.activate();
            }
        }

        else {
            self.content_selection.deactivate();
        }
        Ok(())
    }

    pub fn update(&mut self) -> Result<UIState, Error> {
        match self.rx.recv()? {
            Event::Input(event) => {
                match event.code {
                    KeyCode::Char('q') => Ok(UIState::Terminated),
                    KeyCode::Char('s') => {
                        self.menu_selection = MenuSelection::Stats;
                        Ok(UIState::Render)
                    },
                    KeyCode::Left => {
                        self.menu_selection = self.menu_selection.prev();
                        Ok(UIState::Render)
                    },
                    KeyCode::Right => {
                        self.menu_selection = self.menu_selection.next();
                        Ok(UIState::Render)
                    },
                    KeyCode::Down => {
                        if matches!(self.menu_selection, MenuSelection::Stats) {
                            self.content_selection.increment(self.data.list.len());
                            Ok(UIState::Render)
                        }

                        else {
                            Ok(UIState::Alive)
                        }
                    },
                    KeyCode::Up => {
                        if matches!(self.menu_selection, MenuSelection::Stats) {
                            self.content_selection.decrement(self.data.list.len());
                            Ok(UIState::Render)
                        }

                        else {
                            Ok(UIState::Alive)
                        }
                    },
                    KeyCode::Enter => {
                        if matches!(self.menu_selection, MenuSelection::Quit) {
                            Ok(UIState::Terminated)
                        }
                        else {
                            Ok(UIState::Alive)
                        }
                    },
                    _                  => Ok(UIState::Alive)
                }
            },
            Event::Tick => {
                let terminal_size = self.terminal.size().expect("Getting size of terminal");
                if terminal_size != self.console_size {
                    self.console_size = terminal_size;
                    Ok(UIState::Render)
                } 

                else {
                    Ok(UIState::Alive)
                }
            },
            Event::ForceRender => Ok(UIState::Render)
        }
    }

    pub fn render(&mut self) -> Result<(), Error> {
        self.terminal.draw(|frame| {
            // Create layout elements
            let main_layout = Layout::default().direction(Direction::Vertical).constraints([
                Constraint::Length(3),
                Constraint::Length(3),
                Constraint::Min(3)
            ]).split(frame.size()).to_vec();
            let stats_layout = Layout::default().direction(Direction::Vertical).constraints([
                Constraint::Min(3),
                Constraint::Length(3)
            ]).split(main_layout[2]).to_vec();

            // Elements
            let titel_bar = {
                Paragraph::new("psxreadmap")
                .style(Style::default().fg(Color::White))
                .alignment(Alignment::Center)
                .block(Block::default()
                    .borders(Borders::ALL)
                    .style(Style::default().fg(Color::White))
                    .border_type(BorderType::Plain)
                )    
            };
            let menu_bar = {
                const MENU_TITLES: &'static [&'static str] = &["Stats", "Quit"];

                let menu = MENU_TITLES.iter().map(|t| {
                    let (first, rest) = t.split_at(1);
                    Line::from(vec![
                        Span::styled(first, Style::default().fg(Color::Yellow).add_modifier(Modifier::UNDERLINED),),
                        Span::styled(rest,  Style::default().fg(Color::White)),
                    ])
                }).collect();
                
                Tabs::new(menu).select(self.menu_selection.get_id()).block(
                    Block::default()
                    .title("Menu").borders(Borders::ALL))
                    .style(Style::default().fg(Color::White))
                    .highlight_style(Style::default().bg(Color::LightYellow))
                    .divider(Span::raw("|")
                )
            };

            // Render screen
            frame.render_widget(titel_bar, main_layout[0]);
            frame.render_widget(menu_bar, main_layout[1]);
            match self.menu_selection {
                MenuSelection::Stats => {
                    Self::render_stats(frame, &stats_layout, &self.content_selection, &mut self.data);
                },
                MenuSelection::Quit  => {
                    frame.render_widget(Self::create_exit_message(), main_layout[2]);
                }
            }
        })?;
        Ok(())
    }

    pub fn close(mut self) -> Result<(), Error> {
        crossterm::terminal::disable_raw_mode()?;
        self.terminal.show_cursor()?;
        self.terminal.clear()?;

        Ok(())
    }

    fn render_stats(frame: &mut ConsoleFrame, frames: &Vec<Rect>, content_selection: &ListSelection, data: &mut ConsoleUIData) {
        let content_list = List::new({
            let mut vec = Vec::new();

            for item in &data.list {
                vec.push(ListItem::new(item.as_str()))
            }
            vec
        }).style(Style::default().fg(Color::White))
        .block(Block::default()
            .borders(Borders::ALL)
            .style(Style::default().fg(Color::White))
            .border_type(BorderType::Plain)
        )
        .highlight_style(Style::default().bg(Color::Yellow));
        let     mem_gauge  = Self::create_mem_gauge("Overall memory usage", data.highest_adr, Color::Cyan);
        let mut list_state = ListState::default();
        
        list_state.select(content_selection.selection);

        frame.render_stateful_widget(content_list, frames[0], &mut list_state);
        frame.render_widget(mem_gauge,    frames[1]);
    }

    fn create_mem_gauge<'a>(name: &str, adr: u64, color: Color) -> Gauge<'a> {
        const RAM_TOP:u64 = ConsoleUIData::HIGHEST_RAM_ADDRESS & !ConsoleUIData::RAM_BASE_ADDRESS;

        let adr_masked = adr & !ConsoleUIData::RAM_BASE_ADDRESS;
        let adr_ratio  = (adr_masked as f64)/RAM_TOP as f64;

        Gauge::default().block(Block::default().borders(Borders::ALL)
            .title(format!("{} [0x{:x}/0x{:x}]", name, adr, ConsoleUIData::HIGHEST_RAM_ADDRESS)))
            .gauge_style(Style::default().fg(color).bg(Color::Black).add_modifier(Modifier::BOLD))
            .ratio(adr_ratio).label(format!("{}%", adr_ratio*100.0f64))
    }

    fn create_exit_message<'a>() -> Paragraph<'a> {
        Paragraph::new("Press \"ENTER\" to exit")
            .style(Style::default().fg(Color::White))
            .alignment(Alignment::Center)
            .block(Block::default()
                .borders(Borders::ALL)
                .style(Style::default().fg(Color::White))
                .border_type(BorderType::Plain).title("QUIT")
            )
    }
}

#[derive(Default)]
struct ConsoleUIData {
    highest_adr: u64,
    list:        Vec<String>
}

impl ConsoleUIData {
    pub const RAM_BASE_ADDRESS:u64    = 0x80000000;
    pub const HIGHEST_RAM_ADDRESS:u64 = Self::RAM_BASE_ADDRESS + (2*1024*1024);

    fn _new(memory_map: &MemoryMap) -> Result<ConsoleUIData, Error> {
        let mut ui_data = ConsoleUIData::default();

        ui_data.update(memory_map)?;
        Ok(ui_data)
    }

    pub fn update(&mut self, memory_map: &MemoryMap) -> Result<(), Error> {
        fn get_last_section_end_adr(memory_map: &MemoryMap) -> u64 {
            if let Some(section) = memory_map.sections.last() {
                return section.adr + section.size as u64;
            }

            0u64
        }

        self.highest_adr = get_last_section_end_adr(memory_map);
        self.list = vec![
            String::from("Schwimmflugel 0"),
            String::from("Schwimmflugel 1"),
            String::from("Schwimmflugel 2"),
            String::from("Schwimmflugel 3"),
            String::from("Schwimmflugel 4"),
            String::from("Schwimmflugel 5"),
            String::from("Schwimmflugel 6"),
            String::from("Schwimmflugel 7"),
            ];
        Ok(())
    }
}

pub fn load_memory_map(use_wsl: bool, input: PathBuf) -> Result<MemoryMap, Error> {
    readmap_helper::generate_memory_map(use_wsl, input)
}